library bloc_stream;

import 'dart:async';

import 'package:meta/meta.dart';
import 'package:rxdart/rxdart.dart';

typedef BlocStreamAction<State, B extends BlocStream<State>> = Future<void>
    Function(
  StreamController<State> controller,
  B bloc,
  State currentValue,
);

Stream<State> Function(dynamic) actionToStream<State>(BlocStream<State> bloc) =>
    (action) {
      final controller = StreamController<State>();
      final future = action(controller, bloc, bloc.value) as Future<void>;
      future
          .catchError(controller.addError)
          .whenComplete(() => controller.close());
      return controller.stream;
    };

abstract class BlocStream<State> extends Stream<State> {
  BlocStream(State initialValue) {
    _state$ = BehaviorSubject.seeded(initialValueOverride ?? initialValue);
    _bindActionsToStates();
  }

  final _actionController = StreamController<dynamic>();
  BehaviorSubject<State> _state$;
  StreamSubscription<State> _transformSubscription;

  State get initialValueOverride => null;
  State get value => _state$.value;

  @override
  bool get isBroadcast => _state$.isBroadcast;

  @override
  StreamSubscription<State> listen(
    void Function(State) onData, {
    Function onError,
    void Function() onDone,
    bool cancelOnError,
  }) {
    return _state$.listen(
      onData,
      onError: onError,
      onDone: onDone,
      cancelOnError: cancelOnError,
    );
  }

  /// Add actions to be processed. Implements Sink
  void add<B extends BlocStream<State>>(BlocStreamAction<State, B> action) {
    if (_actionController.isClosed) return;
    _actionController.add(action as dynamic);
  }

  @mustCallSuper
  void willUpdateValue(State current, State next) {}

  /// Transforms a stream of actions to a stream of state.
  /// You can override this to throttle actions for example.
  Stream<State> transformActions(
    Stream<dynamic> actions$,
    Stream<State> Function(BlocStreamAction<State, BlocStream<State>>)
        transform,
  ) =>
      actions$.asyncExpand(transform);

  void _bindActionsToStates() {
    _transformSubscription = transformActions(
      _actionController.stream,
      actionToStream(this),
    ).listen((newState) {
      if (newState == value) return;
      willUpdateValue(value, newState);
      _state$.add(newState);
    }, onError: (error) {
      _state$.addError(error);
    });
  }

  /// Helper method to reduce boilerplate when composing BLoC's together.
  final _closers = <void Function()>[];
  void cancelOnClose(StreamSubscription subscription,
      [void Function() onClose]) {
    _closers.add(() {
      subscription.cancel();
      if (onClose != null) onClose();
    });
  }

  @mustCallSuper
  Future<void> close() {
    _closers.forEach((fn) => fn());
    _transformSubscription.cancel();
    _actionController.close();
    return _state$.close();
  }
}
